(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: plasma.ml 239 2010-06-23 16:49:03Z gerd $ *)

open Printf
open Plasma_rpcapi_aux
open Plasma_util.Operators

let usage() =
  prerr_endline "usage: plasma <mode> [-help|options]";
  prerr_endline "";
  prerr_endline "  available modes:";
  prerr_endline "  - list";
  prerr_endline "  - create";
  prerr_endline "  - delete";
  (* prerr_endline "  - rename"; *)
  prerr_endline "  - mkdir";
  prerr_endline "  - put";
  prerr_endline "  - get";
  prerr_endline "  - fsstat"


let access_cluster esys =
  let cluster = ref None in
  let namenodes = ref [] in
  let rep = ref 0 in
  let debug = ref false in
  let args =
    [ "-cluster", Arg.String (fun s -> cluster := Some s),
      "<name>    Set the cluster name";
      
      "-namenode", Arg.String (fun n -> namenodes := n :: !namenodes),
      "<host>:<port>   Also use this namenode (can be given several times)";

      "-rep", Arg.Set_int rep,
      "<n>   Set replication factor for new files (default 1)";

      "-debug", Arg.Set debug,
      "    Enable debug output";
    ] in
  ( object
      method args = args
      method cluster = !cluster
      method namenodes = !namenodes
      method rep = !rep
      method open_cluster() =
	Netlog.current_logger := 
	  Netlog.channel_logger stderr (if !debug then `Debug else `Info);
	let nn_nodes =
	  if !namenodes = [] then None else Some !namenodes in
	let cfg = Plasma_client_config.get_config
	            ?clustername:!cluster
                    ?nn_nodes () in
	let c = Plasma_client.open_cluster_cc cfg esys in
	c
    end
  )

let list() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let ac = access_cluster esys in
  let one = ref false in
  Arg.parse
    ( ac#args @
	[ "-1", Arg.Set one,
	  "   output one filename per line and nothing else"
	] )
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma list <file> ...";

  let c = ac#open_cluster() in

  let show_file name ii =
    if !one then
      print_endline name
    else
      let rwx n =
	( if n land 4 <> 0 then "r" else "-" ) ^ 
	  ( if n land 2 <> 0 then "w" else "-" ) ^ 
	  ( if n land 1 <> 0 then "x" else "-" ) in
      let bits =
	( match ii.filetype with
	    | `ftype_regular -> "-"
	    | `ftype_directory -> "d"
	    | `ftype_symlink -> "s" 
	) ^
	  rwx (ii.mode lsr 6) ^ 
	  rwx (ii.mode lsr 3) ^ 
	  rwx ii.mode in
      printf "%s %-8s %-8s %12Ld %s %s\n"
	bits
	ii.usergroup.user
	ii.usergroup.group
	ii.eof
	(Netdate.format
	   ~fmt:"%Y-%m-%d %H:%M"
	   (Netdate.create 
	      ~zone:Netdate.localzone
	      (Int64.to_float ii.mtime.tsecs)))
	name
  in

  List.iter
    (fun file ->
       let t = Plasma_client.start c in
       let inode = Plasma_client.lookup t file in
       let ii = Plasma_client.get_inodeinfo t inode in
       ( match ii.filetype with
	   | `ftype_regular | `ftype_symlink ->
	       show_file file ii
	   | `ftype_directory ->
	       if not !one then printf "%s:\n%!" file;
	       let elements = Plasma_client.list_inode t inode in
	       List.iter
		 (fun (elem,elem_inode) ->
		    try
		      let elem_ii =
			Plasma_client.get_inodeinfo t elem_inode in
		      show_file elem elem_ii
		    with
		      | Plasma_client.Plasma_error `enoent -> ()
		 )
		 elements
       );
       Plasma_client.commit t;
    )
    !files



let create() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let ac = access_cluster esys in
  Arg.parse
    ac#args
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma create <file> ...";

  let c = ac#open_cluster () in

  let ii = Plasma_client.regular_ii c 0o666 in
  let t = Plasma_client.start c in
  List.iter
    (fun name ->
       let inode = Plasma_client.create_inode t ii in
       printf "Created inode: %Ld\n%!" inode;
       Plasma_client.link t name inode;
       printf "Linked name: %s -> %Ld\n%!" name inode
    )
    !files;
  Plasma_client.commit t;
  printf "Committed.\n%!"


let mkdir() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let ac = access_cluster esys in
  Arg.parse
    ac#args
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma mkdir <file> ...";

  let c = ac#open_cluster () in

  let ii = Plasma_client.dir_ii c 0o777 in
  let t = Plasma_client.start c in
  List.iter
    (fun name ->
       let inode = Plasma_client.mkdir t name ii in
       printf "Created directory %s -> %Ld\n%!" name inode
    )
    !files;
  Plasma_client.commit t;
  printf "Committed.\n%!"


let delete() =
  let esys = Unixqueue.create_unix_event_system() in
  let files = ref [] in
  let ac = access_cluster esys in

  Arg.parse
    ac#args
    (fun s ->
       files := !files @ [s]
    )
    "usage: plasma delete <file> ...";

  let c = ac#open_cluster () in
  let t = Plasma_client.start c in
  List.iter
    (fun name ->
       Plasma_client.unlink t name;
       printf "Deleted: %s\n%!" name
    )
    !files;
  Plasma_client.commit t


let put() =
  let esys = Unixqueue.create_unix_event_system() in
  let ac = access_cluster esys in

  let plasma_filename = ref "" in
  let plasma_set = ref false in
  let local_filename = ref "" in
  let local_set = ref false in
  let topo = ref `Star in

  Arg.parse
    ( ac#args
      @ [ "-chain", Arg.Unit (fun () -> topo := `Chain),
	  "   Use chain topology instead of star topology"
	]
    )
    (fun s ->
       if not !local_set then (
	 local_filename := s;
	 local_set := true
       )
       else
	 if not !plasma_set then (
	   plasma_filename := s;
	   plasma_set := true
	 )
	 else
	   raise(Arg.Bad("Unexpected arg: " ^ s))
    )
    "usage: plasma put <local_filename> <plasma_filename>";

  if not !plasma_set then
    failwith "No plasma filename given";   
  if not !local_set then
    failwith "No local filename given";

  let fd = Unix.openfile !local_filename [Unix.O_RDONLY] 0 in
  let len = Unix.LargeFile.lseek fd 0L Unix.SEEK_END in
  ignore(Unix.LargeFile.lseek fd 0L Unix.SEEK_SET);
  
  let c = ac#open_cluster () in
  let ii = Plasma_client.regular_ii c 0o666 in
  let t = Plasma_client.start c in
  let inode = Plasma_client.create_file t !plasma_filename ii in
  Plasma_client.commit t;
  Plasma_client.copy_in c inode 0L fd len !topo;
  Unix.close fd;

  printf "Copied: %s -> %s\n%!" !local_filename !plasma_filename


let alt_copy_out esys c inode pos fd len par =
  (* par: number of parallel requests *)

  Plasma_client.configure_buffer c (2*par);

  let t = Plasma_client.start c in
  let blocksize = Plasma_client.blocksize c in
  let blocksizeL = Int64.of_int blocksize in

  let stream_e, signal =
    Plasma_util.signal_engine esys in

  let pool = Netsys_mem.create_pool blocksize in

  let n = ref 0 in
  let p = ref pos in
  let end_pos = Int64.add pos len in

  let rec read_block_e p0 buf r bsize =
    Plasma_client.read_e c inode p0 buf r (bsize-r)
    ++ (fun (l,eof,_) ->
	  if not eof && (r+l) < bsize then
	    read_block_e (Int64.add p0 (Int64.of_int l)) buf (r+l) bsize
	  else
	    eps_e (`Done(r+l,eof)) esys
       )
  in

  let rec fetch() =
    if !n < par then (
      if !p < end_pos then (
	let buf = Netsys_mem.pool_alloc_memory pool in
	let p0 = !p in
	let bsize = Int64.to_int (min blocksizeL (Int64.sub end_pos p0)) in
	let _e =
	  read_block_e p0 (`Memory buf) 0 bsize
	  ++ (fun (l,eof) ->
		ignore(Unix.LargeFile.lseek fd p0 Unix.SEEK_SET);
		ignore(Netsys_mem.mem_write fd buf 0 l);
		decr n;
		fetch();
		eps_e (`Done ()) esys
	     ) in
	incr n;
	p := Int64.add !p blocksizeL;
	fetch()
      )
    );
    if !n=0 then signal(`Done())
  in

  fetch();
  let _e =
    stream_e ++ (fun () -> Plasma_client.commit_e t) in
  Unixqueue.run esys


let get() =
  let esys = Unixqueue.create_unix_event_system() in
  let ac = access_cluster esys in

  let plasma_filename = ref "" in
  let plasma_set = ref false in
  let local_filename = ref "" in
  let local_set = ref false in
  let alt = ref None in

  Arg.parse
    ( ac#args @
	[ "-alt", Arg.Int (fun n -> alt := Some n),
	  "<n>   use the alternate algorithm with a parallelization factor of n"
	]
    )
    (fun s ->
       if not !plasma_set then (
	 plasma_filename := s;
	 plasma_set := true
       )
       else
	 if not !local_set then (
	   local_filename := s;
	   local_set := true
	 )
	 else
	   raise(Arg.Bad("Unexpected arg: " ^ s))
    )
    "usage: plasma get <plasma_filename> <local_filename>";

  if not !plasma_set then
    failwith "No plasma filename given";   
  if not !local_set then
    failwith "No local filename given";

  let fd = 
    Unix.openfile 
      !local_filename [Unix.O_WRONLY; Unix.O_TRUNC; Unix.O_CREAT] 0o666 in
  
  let c = ac#open_cluster () in

  let t = Plasma_client.start c in
  let inode = Plasma_client.lookup t !plasma_filename in
  let ii = Plasma_client.get_inodeinfo t inode in
  Plasma_client.commit t;
  ( match !alt with
      | None ->
	  Plasma_client.copy_out c inode 0L fd ii.eof
      | Some par ->
	  alt_copy_out esys c inode 0L fd ii.eof par
  );
  Unix.close fd;

  printf "Copied: %s -> %s\n%!" !plasma_filename !local_filename


let fsstat() =
  let esys = Unixqueue.create_unix_event_system() in
  let ac = access_cluster esys in

  Arg.parse
    ac#args
    (fun s ->
       raise(Arg.Bad("Unexpected arg: " ^ s))
    )
    "usage: plasma fsstat";

  let c = ac#open_cluster () in
  let st = Plasma_client.fsstat c in
  printf "Total:        %20Ld blocks\n" st.total_blocks;
  printf "Used:         %20Ld blocks\n" st.used_blocks;
  printf "Transitional: %20Ld blocks\n" st.trans_blocks;
  printf "Free:         %20Ld blocks\n"
    (Int64.sub (Int64.sub st.total_blocks st.used_blocks) st.trans_blocks);
  ()

let help() =
  usage()


let dispatch() =
  if Array.length Sys.argv < 2 then
    failwith "Bad usage";

  let mode_str = Sys.argv.(1) in
  incr Arg.current;

  match mode_str with
    | "create" -> create()
    | "delete" -> delete()
    | "mkdir" -> mkdir()
    | "put" -> put()
    | "get" -> get()
    | "list" -> list()
    | "fsstat" -> fsstat()
    | "help" | "-help" | "--help" -> help()
    | _ -> failwith ("Bad usage")

let main () =
  try
    dispatch()
  with
    | Plasma_client.Plasma_error code ->
        prerr_endline ("plasma: " ^ Plasma_util.string_of_errno code);
        exit 2

    | Failure m ->
        prerr_endline ("plasma: " ^ m);
        usage();
        exit 2
          
let () = 
  let ctrl = Gc.get() in
  let ctrl' = { ctrl with 
                  Gc.max_overhead = 1_000_000;  (* Turn compactor off *)
                  Gc.space_overhead = 300;      (* allow more wasted mem *)
                  Gc.minor_heap_size = 65536;   (* bigger minor heap *)
              } in
  Gc.set ctrl';
  main()
